@@ -8,7 +8,7 @@ of this file or portions of this file for uses not covered by the License
 Agreement is not allowed without a written agreement signed by an officer of
 Miva, Inc.
 
-Copyright 1998-2020 Miva, Inc.  All rights reserved.
+Copyright 1998-2022 Miva, Inc.  All rights reserved.
 https://www.miva.com
 
 Prefix         : MMT-COMMAND-
@@ -27,7 +27,9 @@ import merchantapi.abstract
 import merchantapi.authenticator
 import merchantapi.client
 import merchantapi.model
+import merchantapi.logging
 
+from mmt.version import __version__ as version
 from mmt.exceptions import Error, APIRequestError
 from mmt.file import File, BinaryFile
 from mmt.managers.config import ConfigManager
@@ -65,6 +67,11 @@ class Command( abc.ABC ):
 		self._cached_remote					= None
 		self._cached_credential				= None
 
+		self._request_logger				= None
+
+		if args.get( 'debug' ):
+			self._request_logger			= merchantapi.logging.ConsoleLogger( 'stdout' )
+
 	def validate( self ):
 		pass
 
@@ -192,6 +199,8 @@ class Command( abc.ABC ):
 		credential	= self.load_credential( credential_key )
 		client		= self._build_request_client( credential )
 		client.set_option( 'default_store_code', store_code )
+		client.set_global_header( 'User-Agent', f'MMT/{version}' )
+		client.set_logger( self._request_logger )
 
 		if credential.http_basic_auth_username is not None and credential.http_basic_auth_password is not None:
 			encoded = self.base64_encode( f'{credential.http_basic_auth_username}:{credential.http_basic_auth_password}'.encode( 'utf-8' ) ).decode( 'ascii' )
@@ -395,9 +404,6 @@ class Command( abc.ABC ):
 	def build_template_settings_filepath( self, branchtemplateversion: merchantapi.model.BranchTemplateVersion ) -> str:
 		return self._build_template_filepath_lowlevel( branchtemplateversion, '.json' )
 
-	def _build_template_filepath_lowlevel( self, branchtemplateversion: merchantapi.model.BranchTemplateVersion, suffix: str ) -> str:
-		return os.path.join( 'templates', f'{os.path.splitext( self.get_branchtemplateversion_filename( branchtemplateversion ) )[ 0 ]}{suffix}' )
-
 	def build_property_template_filepath( self, branchpropertyversion: merchantapi.model.BranchPropertyVersion ) -> str:
 		return self._build_property_filepath_lowlevel( branchpropertyversion, '.mvt' )
 
@@ -405,28 +411,40 @@ class Command( abc.ABC ):
 		return self._build_property_filepath_lowlevel( branchpropertyversion, '.json' )
 
 	def build_jsresource_settings_filepath( self, branchjavascriptresourceversion: merchantapi.model.BranchJavaScriptResourceVersion ) -> str:
-		return os.path.join( 'js', f'{self.get_branchjavascriptresourceversion_code( branchjavascriptresourceversion )}.json' )
+		return self._build_jsresource_filepath_lowlevel( branchjavascriptresourceversion, '.json' )
 
 	def build_jsresource_local_filepath( self, branchjavascriptresourceversion: merchantapi.model.BranchJavaScriptResourceVersion ) -> str:
-		return os.path.join( 'js', f'{self.get_branchjavascriptresourceversion_code( branchjavascriptresourceversion )}.js' )
+		return self._build_jsresource_filepath_lowlevel( branchjavascriptresourceversion, '.js' )
 
 	def build_jsresource_inline_filepath( self, branchjavascriptresourceversion: merchantapi.model.BranchJavaScriptResourceVersion ) -> str:
-		return os.path.join( 'js', f'{self.get_branchjavascriptresourceversion_code( branchjavascriptresourceversion )}.mvt' )
+		return self._build_jsresource_filepath_lowlevel( branchjavascriptresourceversion, '.mvt' )
 
 	def build_cssresource_settings_filepath( self, branchcssresourceversion: merchantapi.model.BranchCSSResourceVersion ) -> str:
-		return os.path.join( 'css', f'{self.get_branchcssresourceversion_code( branchcssresourceversion )}.json' )
+		return self._build_cssresource_filepath_lowlevel( branchcssresourceversion, '.json' )
 
 	def build_cssresource_local_filepath( self, branchcssresourceversion: merchantapi.model.BranchCSSResourceVersion ) -> str:
-		return os.path.join( 'css', f'{self.get_branchcssresourceversion_code( branchcssresourceversion )}.css' )
+		return self._build_cssresource_filepath_lowlevel( branchcssresourceversion, '.css' )
 
 	def build_cssresource_inline_filepath( self, branchcssresourceversion: merchantapi.model.BranchCSSResourceVersion ) -> str:
-		return os.path.join( 'css', f'{self.get_branchcssresourceversion_code( branchcssresourceversion )}.mvt' )
+		return self._build_cssresource_filepath_lowlevel( branchcssresourceversion, '.mvt' )
 
 	def build_resourcegroup_settings_filepath( self, resourcegroup: merchantapi.model.ResourceGroup ) -> str:
-		return os.path.join( 'resourcegroups', f'{self.get_resourcegroup_code( resourcegroup )}.json' )
+		return self._build_resourcegroup_filepath_lowlevel( resourcegroup, '.json' )
+
+	def _build_template_filepath_lowlevel( self, branchtemplateversion: merchantapi.model.BranchTemplateVersion, suffix: str ) -> str:
+		return self._build_filepath_lowlevel( 'templates', os.path.splitext( self.get_branchtemplateversion_filename( branchtemplateversion ) )[ 0 ], suffix )
+
+	def _build_jsresource_filepath_lowlevel( self, branchjavascriptresourceversion: merchantapi.model.BranchJavaScriptResourceVersion, suffix: str ) -> str:
+		return self._build_filepath_lowlevel( 'js', self.get_branchjavascriptresourceversion_code( branchjavascriptresourceversion ), suffix )
+
+	def _build_cssresource_filepath_lowlevel( self, branchcssresourceversion: merchantapi.model.BranchCSSResourceVersion, suffix: str ) -> str:
+		return self._build_filepath_lowlevel( 'css', self.get_branchcssresourceversion_code( branchcssresourceversion ), suffix )
 
 	def _build_property_filepath_lowlevel( self, branchpropertyversion: merchantapi.model.BranchPropertyVersion, suffix: str ) -> str:
-		return os.path.join( self._build_property_path( branchpropertyversion ), f'{self.get_branchpropertyversion_code( branchpropertyversion )}{suffix}' )
+		return self._build_filepath_lowlevel( self._build_property_path( branchpropertyversion ), self.get_branchpropertyversion_code( branchpropertyversion ), suffix )
+
+	def _build_resourcegroup_filepath_lowlevel( self, resourcegroup: merchantapi.model.ResourceGroup, suffix: str ) -> str:
+		return self._build_filepath_lowlevel( 'resourcegroups', self.get_resourcegroup_code( resourcegroup ), suffix )
 
 	def _build_property_path( self, branchpropertyversion: merchantapi.model.BranchPropertyVersion ) -> str:
 		if len( branchpropertyversion.get_code() ):
@@ -440,6 +458,16 @@ class Command( abc.ABC ):
 
 		return os.path.join( 'properties', os.sep.join( directories ) )
 
+	def _build_filepath_lowlevel( self, parent_directory: str, filename_no_ext: str, suffix: str ) -> str:
+		return os.path.join( parent_directory, f'{self._os_safe_filename_no_ext( filename_no_ext )}{suffix}' )
+
+	def _os_safe_filename_no_ext( self, filename_no_ext: str ) -> str:
+		for character in filename_no_ext:
+			if not character.isalnum() and not character in [ '-', '_', ' ' ]:
+				return hashlib.md5( filename_no_ext.encode( 'utf-8' ) ).hexdigest()
+
+		return filename_no_ext
+
 	# JS resource file helper functions
 
 	def build_jsresource_settings( self, branchjavascriptresourceversion: merchantapi.model.BranchJavaScriptResourceVersion ) -> dict:
