@@ -8,7 +8,7 @@ of this file or portions of this file for uses not covered by the License
 Agreement is not allowed without a written agreement signed by an officer of
 Miva, Inc.
 
-Copyright 1998-2020 Miva, Inc.  All rights reserved.
+Copyright 1998-2021 Miva, Inc.  All rights reserved.
 https://www.miva.com
 
 Prefix         : MMT-COMMAND-PULL-
@@ -22,7 +22,7 @@ import merchantapi.request
 
 from mmt.exceptions import Error, ErrorList
 from mmt.commands import ConfiguredCommand
-from mmt.metadata.state import StateMetadataEntryFile, StateMetadataTemplateFile, StateMetadataTemplateSettingsFile, StateMetadataPropertyTemplateFile, StateMetadataPropertySettingsFile, StateMetadataJSResourceJSFile, StateMetadataJSResourceTemplateFile, StateMetadataJSResourceSettingsFile, StateMetadataCSSResourceCSSFile, StateMetadataCSSResourceTemplateFile, StateMetadataCSSResourceSettingsFile, StateMetadataResourceGroupSettingsFile
+from mmt.metadata.state import StateMetadataEntryFile, StateMetadataBaseTemplateFile, StateMetadataTemplateFile, StateMetadataTemplateSettingsFile, StateMetadataBasePropertyFile, StateMetadataPropertyTemplateFile, StateMetadataPropertySettingsFile, StateMetadataBaseJSResourceFile, StateMetadataJSResourceJSFile, StateMetadataJSResourceTemplateFile, StateMetadataJSResourceSettingsFile, StateMetadataBaseCSSResourceFile, StateMetadataCSSResourceCSSFile, StateMetadataCSSResourceTemplateFile, StateMetadataCSSResourceSettingsFile, StateMetadataBaseResourceGroupFile, StateMetadataResourceGroupSettingsFile
 from mmt.file import File, BinaryFile
 
 
@@ -30,22 +30,22 @@ class PullCommand( ConfiguredCommand ):
 	def __init__( self, *args, **kwargs ):
 		super().__init__( *args, **kwargs )
 
-		self._added_files		= []
-		self._updated_files		= []
-		self._deleted_files		= []
+		self._added_files	= []
+		self._updated_files	= []
+		self._deleted_files	= []
 
 	def initialize( self ):
 		self.ensure_toplevel_directories_exist()
 
 	def run( self ):
-		filepaths				= set( filepath for filepath in self.args.get( 'filepaths' ) )
-		modified_files			= self.load_modified_files( filtered_filepaths = filepaths )
-		out_of_sync_files		= []
-		template_filenames		= []
-		property_types_codes	= []
-		jsresource_codes		= []
-		cssresource_codes		= []
-		resourcegroup_codes		= []
+		filepaths					= set( filepath for filepath in self.args.get( 'filepaths' ) )
+		modified_files				= self.load_modified_files( filtered_filepaths = filepaths )
+		out_of_sync_files			= []
+		template_state_files		= []
+		property_state_files		= []
+		jsresource_state_files		= []
+		cssresource_state_files		= []
+		resourcegroup_state_files	= []
 
 		# Check to see if any of the locally modified files are explicitly being updated
 		if not self.args.get( 'force' ):
@@ -64,31 +64,53 @@ class PullCommand( ConfiguredCommand ):
 
 			if state_file is not None:
 				if state_file.is_template():
-					template_filenames.append( state_file.template_filename )
+					template_state_files.append( state_file )
 				elif state_file.is_property():
-					property_types_codes.append( ( state_file.property_type, state_file.property_code ) )
+					property_state_files.append( state_file )
 				elif state_file.is_jsresource():
-					jsresource_codes.append( state_file.jsresource_code )
+					jsresource_state_files.append( state_file )
 				elif state_file.is_cssresource():
-					cssresource_codes.append( state_file.cssresource_code )
+					cssresource_state_files.append( state_file )
 				elif state_file.is_resourcegroup():
-					resourcegroup_codes.append( state_file.resourcegroup_code )
+					resourcegroup_state_files.append( state_file )
 
 		changeset							= self._load_changeset( self.args.get( 'c', 0 ) )
-		branchtemplateversions				= self._listload_branchtemplateversions( changeset, template_filenames )
-		branchpropertyversions				= self._listload_branchpropertyversions( changeset, property_types_codes )
-		branchjavascriptresourceversions	= self._listload_branchjavascriptresourceversions( changeset, jsresource_codes )
-		branchcssresourceversions			= self._listload_branchcssresourceversions( changeset, cssresource_codes )
-		resourcegroups						= self._listload_resourcegroups( changeset, resourcegroup_codes )
+		branchtemplateversions				= []
+		branchpropertyversions				= []
+		branchjavascriptresourceversions	= []
+		branchcssresourceversions			= []
+		resourcegroups						= []
 
-		modified_template_files				= self.group_modified_template_files( modified_files )
-		modified_property_files				= self.group_modified_property_files( modified_files )
-		modified_jsresource_files			= self.group_modified_jsresource_files( modified_files )
-		modified_cssresource_files			= self.group_modified_cssresource_files( modified_files )
-		modified_resourcegroup_files		= self.group_modified_resourcegroup_files( modified_files )
+		if len( filepaths ) == 0:
+			branchtemplateversions					= self._listload_branchtemplateversions( changeset )
+			branchpropertyversions					= self._listload_branchpropertyversions( changeset )
+			branchjavascriptresourceversions		= self._listload_branchjavascriptresourceversions( changeset )
+			branchcssresourceversions				= self._listload_branchcssresourceversions( changeset )
+			resourcegroups							= self._listload_resourcegroups( changeset )
+		else:
+			if len( template_state_files ) > 0:
+				branchtemplateversions				= self._listload_branchtemplateversions( changeset, template_state_files )
+
+			if len( property_state_files ) > 0:
+				branchpropertyversions				= self._listload_branchpropertyversions( changeset, property_state_files )
+
+			if len( jsresource_state_files ) > 0:
+				branchjavascriptresourceversions	= self._listload_branchjavascriptresourceversions( changeset, jsresource_state_files )
+
+			if len( cssresource_state_files ) > 0:
+				branchcssresourceversions			= self._listload_branchcssresourceversions( changeset, cssresource_state_files )
+
+			if len( resourcegroup_state_files ) > 0:
+				resourcegroups						= self._listload_resourcegroups( changeset, resourcegroup_state_files )
 
 		# Pull the source from the server and see if any of locally modified files are out-of-sync
 		if not self.args.get( 'force' ):
+			modified_template_files			= self.group_modified_template_files( modified_files )
+			modified_property_files			= self.group_modified_property_files( modified_files )
+			modified_jsresource_files		= self.group_modified_jsresource_files( modified_files )
+			modified_cssresource_files		= self.group_modified_cssresource_files( modified_files )
+			modified_resourcegroup_files	= self.group_modified_resourcegroup_files( modified_files )
+
 			for state_file, file in modified_template_files:
 				for branchtemplateversion in branchtemplateversions:
 					if self.template_equals_branchtemplateversion( state_file, branchtemplateversion ):
@@ -263,12 +285,12 @@ class PullCommand( ConfiguredCommand ):
 
 		return response.get_changesets()[ 0 ]
 
-	def _listload_branchtemplateversions( self, changeset: merchantapi.model.Changeset, template_filenames: typing.List[ str ] ) -> typing.List[ merchantapi.model.BranchTemplateVersion ]:
+	def _listload_branchtemplateversions( self, changeset: merchantapi.model.Changeset, template_state_files: typing.Optional[ typing.List[ StateMetadataBaseTemplateFile ] ] = None ) -> typing.List[ merchantapi.model.BranchTemplateVersion ]:
 		request = merchantapi.request.BranchTemplateVersionListLoadQuery()
 		filters = request.filter_expression()
 
-		if len( template_filenames ) > 0:
-			filters.is_in( 'filename', template_filenames )
+		if template_state_files is not None:
+			filters.is_in( 'filename', [ state_file.template_filename for state_file in template_state_files ] )
 		else:
 			filters.equal( 'prop_id', 0 )
 
@@ -282,55 +304,88 @@ class PullCommand( ConfiguredCommand ):
 
 		return self.send_request_base64( request ).get_branch_template_versions()
 
-	def _listload_branchpropertyversions( self, changeset: merchantapi.model.Changeset, property_types_codes: typing.List[ typing.Tuple[ str, str ] ] ) -> typing.List[ merchantapi.model.BranchPropertyVersion ]:
+	def _listload_branchpropertyversions( self, changeset: merchantapi.model.Changeset, property_state_files: typing.Optional[ typing.List[ StateMetadataBasePropertyFile ] ] = None ) -> typing.List[ merchantapi.model.BranchPropertyVersion ]:
 		request = merchantapi.request.BranchPropertyVersionListLoadQuery()
 		filters = request.filter_expression()
 
-		if len( property_types_codes ) > 0:
-			for property_type, property_code in property_types_codes:
-				filters.or_x( request.filter_expression().equal( 'type', property_type ).and_equal( 'code', property_code ) )
-		else:
+		if property_state_files is None:
 			if self.configmanager.ignore_unsynced_properties:
 				filters.and_is_true( 'sync' )
+		else:
+			for state_file in property_state_files:
+				filter_expression = request.filter_expression().equal( 'type', state_file.property_type )
+
+				if state_file.property_group in [ 'product', 'category' ]:
+					filter_expression.and_is_null( 'code' )
+				else:
+					filter_expression.and_equal( 'code', state_file.property_code )
+
+				filters.or_x( filter_expression )
 
 		request.set_filters( filters )
 		request.set_changeset_id( changeset.get_id() )
 		request.set_branch_id( self.configmanager.branch_id )
 		request.set_on_demand_columns( [ 'source', 'settings', 'category', 'product' ] )
 
-		return self.send_request_base64( request ).get_branch_property_versions()
+		branchpropertyversions = self.send_request_base64( request ).get_branch_property_versions()
 
-	def _listload_branchjavascriptresourceversions( self, changeset: merchantapi.model.Changeset, jsresource_codes: typing.List[ str ] ) -> typing.List[ merchantapi.model.BranchJavaScriptResourceVersion ]:
+		# If no filter is being applied return all elements, otherwise manually apply
+		# an additional filter to ensure only properties requested are returned.  This
+		# is necessary since product / category properties cannot have their code values
+		# searched as part of the request filter
+
+		if property_state_files is None:
+			return branchpropertyversions
+		else:
+			filtered_branchpropertyversions = []
+
+			for branchpropertyversion in branchpropertyversions:
+				for state_file in property_state_files:
+					if self.property_equals_branchpropertyversion( state_file, branchpropertyversion ):
+						filtered_branchpropertyversions.append( branchpropertyversion )
+
+			return filtered_branchpropertyversions
+
+	def _listload_branchjavascriptresourceversions( self, changeset: merchantapi.model.Changeset, jsresource_state_files: typing.Optional[ typing.List[ StateMetadataBaseJSResourceFile ] ] = None ) -> typing.List[ merchantapi.model.BranchJavaScriptResourceVersion ]:
 		request = merchantapi.request.BranchJavaScriptResourceVersionListLoadQuery()
+		filters = request.filter_expression()
+
+		if jsresource_state_files is not None:
+			filters.is_in( 'code', [ state_file.jsresource_code for state_file in jsresource_state_files ] )
+
+		request.set_filters( filters )
 		request.set_changeset_id( changeset.get_id() )
 		request.set_branch_id( self.configmanager.branch_id )
 		request.set_on_demand_columns( [ 'source', 'linkedpages', 'linkedresources' ] )
 
-		if len( jsresource_codes ) > 0:
-			request.set_filters( request.filter_expression().is_in( 'code', jsresource_codes ) )
-
 		return self.send_request_base64( request ).get_branch_java_script_resource_versions()
 
-	def _listload_branchcssresourceversions( self, changeset: merchantapi.model.Changeset, cssresource_codes: typing.List[ str ] ) -> typing.List[ merchantapi.model.BranchCSSResourceVersion ]:
+	def _listload_branchcssresourceversions( self, changeset: merchantapi.model.Changeset, cssresource_state_files: typing.Optional[ typing.List[ StateMetadataBaseCSSResourceFile ] ] = None ) -> typing.List[ merchantapi.model.BranchCSSResourceVersion ]:
 		request = merchantapi.request.BranchCSSResourceVersionListLoadQuery()
+		filters	= request.filter_expression()
+
+		if cssresource_state_files is not None:
+			filters.is_in( 'code', [ state_file.cssresource_code for state_file in cssresource_state_files ] )
+
+		request.set_filters( filters )
 		request.set_changeset_id( changeset.get_id() )
 		request.set_branch_id( self.configmanager.branch_id )
 		request.set_on_demand_columns( [ 'source', 'linkedpages', 'linkedresources' ] )
 
-		if len( cssresource_codes ) > 0:
-			request.set_filters( request.filter_expression().is_in( 'code', cssresource_codes ) )
-
 		return self.send_request_base64( request ).get_branch_css_resource_versions()
 
-	def _listload_resourcegroups( self, changeset: merchantapi.model.Changeset, resourcegroup_codes: typing.List[ str ] ) -> typing.List[ merchantapi.model.ResourceGroup ]:
+	def _listload_resourcegroups( self, changeset: merchantapi.model.Changeset, resourcegroup_state_files: typing.Optional[ typing.List[ StateMetadataBaseResourceGroupFile ] ] = None ) -> typing.List[ merchantapi.model.ResourceGroup ]:
 		request = merchantapi.request.ResourceGroupListLoadQuery()
+		filters	= request.filter_expression()
+
+		if resourcegroup_state_files is not None:
+			filters.is_in( 'code', [ state_file.resourcegroup_code for state_file in resourcegroup_state_files ] )
+
+		request.set_filters( filters )
 		request.set_changeset_id( changeset.get_id() )
 		request.set_branch_id( self.configmanager.branch_id )
 		request.set_on_demand_columns( [ 'linkedjavascriptresources', 'linkedcssresources' ] )
 
-		if len( resourcegroup_codes ) > 0:
-			request.set_filters( request.filter_expression().is_in( 'code', resourcegroup_codes ) )
-
 		return self.send_request( request ).get_resource_groups()
 
 	def _process_branchtemplateversion( self, branchtemplateversion: merchantapi.model.BranchTemplateVersion ):
